from logging import getLogger
from typing import Any

from pydantic import BaseModel, model_validator
from typing_extensions import Self

from .api_ret.price_history import PriceHistoryAPIRet
from .api_ret.product_search import ProductSearchAPIRet

logger = getLogger(__name__)


class BaseToolResponse(BaseModel):
    def slim_dump(self) -> str:
        return self.model_dump_json(exclude_none=True, exclude_defaults=True)


class ProductSearchToolResponse(BaseToolResponse):
    product_search_result: ProductSearchAPIRet
    reason: str | None = None
    display_rules: str | None = None

    @model_validator(mode="after")
    def post_init(self) -> Self:
        if len(self.product_search_result.list) == 0:
            self.reason = """
No results found. Possible reasons:
1. This search is much more complex than a simple product search.
2. The user is asking things related to product specifications.

If the problems might be related to the points listed above,
please use the 'spec_search' tool and try again.
            """
            return self

        # add display rules if result is not empty
        else:
            self.display_rules = """
As a product researcher, you need to find the most relavent product and present them in utmost detail.
Without following the rules listed bellow, the output will become useless, you must follow the rules before responding to the user.
All rules must be followed strictly.

Here are a list of rules you must follow:
Rule 1: Product image must be included when available, url is located in each object inside 'specs.images' field.
Rule 2: If no avaliable image exist, ignore the image field completely, don't even write anything image related for that single product.
Rule 3: Product urls must be included so that the user can by the product with a simple click if possible.
Rule 4: Display more then one relavent product if possible, having multiple choices is a good thing.
            """

        return self


class PriceHisotryGraphToolResponse(BaseToolResponse):
    price_history_graph: str
    display_rules: str = """
This is a markdown image link, it must be included in the final output. 
    """


class PriceHistoryToolResponse(BaseToolResponse):
    price_history_description: PriceHistoryAPIRet
    price_history_graph: str
    display_rules: str = """
Field explanation:
'price_history_description': includes detailed price history info.
'price_history_graph': includes a markdown image link used for visualizing price history.

Here are a list of rules you must follow:
Rule 1: Both 'price_history_description' and the link provided at 'price_history_graph' field must be included in the output.
Rule 2: Product url must be included, it can be found in 'price_history_description'
    """


class SpecIndexesToolResponse(BaseToolResponse):
    indexes: list[str]


class SpecMappingToolResponse(BaseToolResponse):
    mappings: dict[str, dict[str, Any]]
    example_document: dict[str, dict[str, Any]]
    note: str = """
Specifications are under the 'specs' field
Example fields paths:
- specs.physical_specs.weight
- specs.technical_specs.water_resistance.depth
- specs.sensors.gyroscope
"""


class SpecSearchToolResponse(BaseToolResponse):
    hits: list[dict[str, Any]]
    reason: str | None = None
    display_rules: str | None = None

    @model_validator(mode="after")
    def post_init(self) -> Self:
        # remove documents with status == deleted
        cleaned_hits: list[dict[str, Any]] = []
        for hit in self.hits:
            if hit["_source"].get("status", None) == "deleted":
                continue
            else:
                cleaned_hits.append(hit)

        logger.debug(
            "original hits: %s, cleaned_hits: %s", len(self.hits), len(cleaned_hits)
        )

        self.hits = cleaned_hits

        if len(self.hits) == 0:
            self.reason = """
No results found. Try:
1. For complex product searches → Use 'spec_search' tool
2. For specification-related queries → Use 'spec_search' tool
            """

        else:
            self.display_rules = """
Display product images (from 'specs.images.url' field) when available; omit image mentions if unavailable
            """

        return self



import json
from logging import getLogger
from typing import Annotated, Any
from mcp.server.fastmcp import Context
from pydantic import Field

from biggo_mcp_server.types.setting import BigGoMCPSetting
from ..types.responses import (
    SpecIndexesToolResponse,
    SpecMappingToolResponse,
    SpecSearchToolResponse,
)
from ..lib.utils import get_setting
from ..services.spec_search import SpecSearchService

logger = getLogger(__name__)


async def spec_indexes(
    ctx: Context,
) -> Annotated[str, Field(description="List of Elasticsearch indexes")]:
    """Elasticsearch Indexes for Product Specification.

    It is REQUIRED to use this tool first before running any specification search.
    """
    logger.info("spec indexes")
    setting = get_setting(ctx)
    service = SpecSearchService(setting)
    async with service.session():
        indexes = await service.spec_indexes()
    return SpecIndexesToolResponse(indexes=indexes).slim_dump()


async def spec_mapping(
    ctx: Context,
    index: Annotated[
        str,
        Field(
            description="""
                          Elasticsearch index

                          Steps to obtain this argument.
                          1. Use 'spec_indexes' tool to get the list of indexes
                          2. Choose the most relevant index
                          """
        ),
    ],
) -> Annotated[
    str, Field(description="Elasticsearch mappings plus an example document")
]:
    """Elasticsearch Mapping For Product Specification.

    Use this tool after you have the index, and need to know the mapping in order to query the index.
    Available indexes can be obtained by using the 'spec_indexes' tool.
    """
    logger.info("spec mapping, index: %s", index)
    setting = get_setting(ctx)
    service = SpecSearchService(setting)
    try:
        async with service.session():
            mapping = await service.spec_mapping(index)
    except Exception as e:
        raise Exception(
            f"You must know the available indexes before using this tool. Error: {e}"
        )
    return SpecMappingToolResponse(
        mappings=mapping.mappings, example_document=mapping.example_document
    ).slim_dump()


async def spec_search(
    ctx: Context,
    index: Annotated[
        str,
        Field(
            description="""
                          Elasticsearch index

                          Steps to obtain this argument.
                          1. Use 'spec_indexes' tool to get the list of indexes
                          2. Choose the most relevant index
                          """
        ),
    ],
    elasticsearch_query: Annotated[
        str | dict[str, Any],
        Field(
            description=f"""
              Elasticsearch(ES8) query tool

              Required steps:
              1. First use 'spec_mapping' to get index mapping
              2. Result limit ≤ 10
              3. Must exclude documents with 'status'='deleted'
              4. Must include documents with 'region'='{BigGoMCPSetting().region.value.lower()}'

              When to sort:
              - For optimal performance (e.g., most energy-efficient refrigerator) → sort by power consumption
              - For specific attribute extremes (e.g., smallest refrigerator) → sort by height
              - When only querying products matching specific parameters (e.g., phones with 16GB RAM) → no sorting needed

              Note: All spec fields are under the 'spec' key, must add 'spec' prefix when querying
              Example paths: specs.physical_specs.weight, specs.technical_specs.water_resistance.depth
              """,
            examples=[
                {
                    "query": {
                        "bool": {
                            "must_not": [{"match": {"status": "deleted"}}],
                            "must": [
                                {
                                    "range": {
                                        "specs.physical_specifications.dimensions.height": {
                                            "gte": 1321,
                                            "lte": 2321,
                                        }
                                    }
                                },
                                {"match": {"region": "tw"}},
                            ],
                        }
                    },
                    "size": 5,
                    "sort": [
                        {"specs.physical_specifications.dimensions.height": "asc"}
                    ],
                }
            ],
        ),
    ],
) -> str:
    """Product Specification Search.

    Index mapping must be aquired before using this tool.
    Use 'spec_mapping' tool to get the mapping of the index.
    """
    logger.info("spec search, index: %s, query: %s", index, elasticsearch_query)

    setting = get_setting(ctx)
    service = SpecSearchService(setting)
    try:
        async with service.session():
            if isinstance(elasticsearch_query, str):
                query: dict[str, Any] = json.loads(elasticsearch_query)
            else:
                query = elasticsearch_query
            hits = await service.search(index, query)
    except Exception as e:
        raise Exception(
            f"You must know the mapping of the index before using this tool. Error: {e}"
        )
    return SpecSearchToolResponse(hits=hits).slim_dump()



from contextlib import contextmanager
import pytest
import sys
from os import environ
from pathlib import Path
from dotenv import load_dotenv
from biggo_mcp_server.types.setting import BigGoMCPSetting


def load_env_files():
    """載入環境變數，相容所有 Python 3.10+ 版本"""
    test_env_paths = [
        Path(__file__).parent / ".env.test",  # test/.env.test
        Path(__file__).parent.parent / ".env.test",  # .env.test
    ]

    for env_path in test_env_paths:
        if env_path.is_file():  # 使用 is_file() 而不是 exists()
            try:
                # Python 3.10+ 相容性處理
                env_path_str = (
                    str(env_path)
                    if sys.version_info >= (3, 11)
                    else env_path.absolute().as_posix()
                )
                load_dotenv(env_path_str)
                return True
            except Exception as e:
                print(f"Warning: Failed to load {env_path}: {e}", file=sys.stderr)
    return False


# 確保環境變數被載入
load_env_files()


@pytest.fixture
def setting():
    client_id = environ.get("BIGGO_MCP_SERVER_CLIENT_ID")
    client_secret = environ.get("BIGGO_MCP_SERVER_CLIENT_SECRET")

    if not client_id or not client_secret:
        raise ValueError(
            "Environment variables BIGGO_MCP_SERVER_CLIENT_ID and BIGGO_MCP_SERVER_CLIENT_SECRET must be set. "
            "Please check if .env.test exists and contains the correct values."
        )

    setting = BigGoMCPSetting(
        client_id=client_id,
        client_secret=client_secret,
    )
    return setting


@contextmanager
def set_temp_env(env_dict: dict[str, str]):
    for k, v in env_dict.items():
        environ[k] = v
    yield
    for k, _ in env_dict.items():
        environ.pop(k)



from __future__ import annotations

from typing import Any, List

from pydantic import BaseModel, Field

from ..setting import Domains


class Multiple(BaseModel):
    min_price: float = 0
    max_price: float = 0
    model_id: str = ""
    title: str = ""
    current_price: float = 0


class Store(BaseModel):
    # image: str
    # link: str
    name: str = ""
    discount_info: str = ""
    # rate_desc: str | None = None
    # is_cashback: bool


class Shop(BaseModel):
    name: str = ""
    username: str = ""
    uid: str = ""
    location: str = ""
    # seller_credit: int | None = None


class ListItem(BaseModel):
    nindex: str = ""
    oid: str = ""
    # subscribe_id: Any
    history_id: str = ""
    item_id: str = ""
    # is_ad: bool
    # is_group: bool
    # is_offline: bool
    # is_notfound: bool
    # is_expired: bool
    # is_adult: bool
    # is_multiple_product: bool
    # is_subscribe: bool
    title: str = ""

    # purl: str | None = None
    affurl: str | None = ""
    url: str = ""

    image: str = ""
    # gallery_count: int
    # origin_image: str
    cata: List[Any] = Field(default_factory=list)
    symbol: str = ""
    currency: str = ""
    multiple: Multiple = Field(default_factory=Multiple)
    price: float = 0
    price_range_min: Any | None = None
    price_range_max: Any | None = None
    count_result_store: Any | None = None
    count_result_product: Any | None = None
    store: Store = Field(default_factory=Store)
    has_shop: bool | None = None
    shop: Shop = Field(default_factory=Shop)
    price_diff_real: float = 0
    product_nindex_price: List[Any] = Field(default_factory=list)
    # subscribe_tags: List
    # subscribe_time: Any


class BiggoCItem(BaseModel):
    key: str = ""
    value: str = ""
    time: int | None = None


class ProductSearchAPIRet(BaseModel):
    # result: bool
    # total: int
    # total_page: int
    # pure_total: int
    # ec_count: int
    # mall_count: int
    # bid_count: int
    # size: int
    # took: int
    # is_shop: bool
    # is_suggest_query: bool
    # is_ypa: bool
    # is_adsense: bool
    # q_suggest: str
    # arr_suggest: List[str]
    # offline_count: int
    # spam_count: int
    # promo: List
    # filter: Dict[str, Any]
    # top_ad_count: int
    # group: Any
    # recommend_group: List
    list: List[ListItem] = Field(default_factory=list)
    # biggo_c: List[BiggoCItem]
    low_price: float = 0
    high_price: float = 0

    def generate_r_link(self, domain: Domains):
        for product in self.list:
            product.url = f"https://{domain.value}{product.affurl}"
            product.affurl = None

    def get_all_urls(self) -> set[str]:
        product_urls = {product.url for product in self.list}
        image_urls = {product.image for product in self.list}
        return product_urls | image_urls

    def replace_urls(self, url_map: dict[str, str]):
        """Replace urls in the list with the new urls

        Args:
            url_map (dict[str, str]): The map of old urls to new urls
        """
        for product in self.list:
            if product.url in url_map:
                product.url = url_map[product.url]
            if product.image in url_map:
                product.image = url_map[product.image]



from asyncio import Event, Task, create_task, CancelledError
from logging import getLogger
from typing import Any
from mcp.server.fastmcp import FastMCP
from ..types.setting import BigGoMCPSetting
import os
import signal

logger = getLogger(__name__)


class BigGoMCPServer(FastMCP):
    def __init__(self, setting: BigGoMCPSetting):
        super().__init__("BigGo MCP Server")
        self._biggo_setting = setting
        self.settings.port = setting.sse_port
        self._end_event: Event = Event()
        self._bg: list[Task[Any]] = []

    @property
    def biggo_setting(self) -> BigGoMCPSetting:
        return self._biggo_setting

    async def _run_stdio_async(self) -> None:
        logger.info("Start stdio BigGo MCP Server")
        try:
            return await self.run_stdio_async()
        except (KeyboardInterrupt, CancelledError) as ex:
            logger.warning(ex)
        except Exception:
            logger.exception("[stdio] somthing is wrong")
        finally:
            self._end_event.set()

    async def _run_sse_async(self) -> None:
        logger.info("Start SSE BigGo MCP Server")
        try:
            await super().run_sse_async()
        except (KeyboardInterrupt, CancelledError) as ex:
            logger.warning(ex)
        except Exception:
            logger.exception("[SSE] somthing is wrong")
        finally:
            self._end_event.set()

    async def start(self):
        stdio_bg = create_task(self._run_stdio_async(), name="stdio-bg")
        self._bg.append(stdio_bg)

        if self.biggo_setting.server_type == "sse":
            sse_bg = create_task(self._run_sse_async(), name="sse-bg")
            self._bg.append(sse_bg)

    async def _cleanup(self):
        for task in self._bg:
            logger.info("cancel task: %s", task.get_name())
            task.cancel()

    async def wait_finish(self):
        await self._end_event.wait()
        logger.info("Server shutting down")
        await self._cleanup()
        logger.info("Server shutdown complete")
        os.kill(os.getpid(), signal.SIGKILL)



from __future__ import annotations


from pydantic import BaseModel, RootModel


class Days(BaseModel):
    days: int
    max_price: float
    min_price: float
    up_times: int
    down_times: int
    last_continue_times: int
    last_increase_status: bool


class Statistics(BaseModel):
    days730: Days | None = None
    days365: Days | None = None
    days180: Days | None = None
    days90: Days | None = None
    days30: Days | None = None
    days7: Days | None = None


class PriceHistoryAttributes(BaseModel):
    symbol: str
    currency: str
    # nindex: str
    # oid: str
    current_price: float
    # datetime_format: str
    # price_history: List[PriceHistoryItem]
    # purl: str | None = None
    url: str | None = None
    title: str
    nindex_name: str
    # icon: str
    statistics: Statistics
    state: str


class PriceHistoryAPIRet(RootModel[dict[str, PriceHistoryAttributes]]):
    root: dict[str, PriceHistoryAttributes]

    def get_all_urls(self) -> set[str]:
        return {item.url for item in self.root.values() if item.url is not None}

    def replace_urls(self, url_map: dict[str, str]):
        for item in self.root.values():
            if item.url is not None and item.url in url_map:
                item.url = url_map[item.url]



from enum import Enum
from typing import Any, Literal
from pydantic_settings import BaseSettings, SettingsConfigDict
from pydantic import field_validator


class GraphLanguage(str, Enum):
    TW = "tw"
    EN = "en"


class LogLevel(str, Enum):
    DEBUG = "DEBUG"
    INFO = "INFO"
    WARNING = "WARNING"
    ERROR = "ERROR"
    CRITICAL = "CRITICAL"


class Regions(str, Enum):
    ID = "ID"
    VN = "VN"
    TH = "TH"
    PH = "PH"
    MY = "MY"
    IN = "IN"
    US = "US"
    TW = "TW"
    HK = "HK"
    JP = "JP"
    SG = "SG"


class Domains(str, Enum):
    ID = "biggo.id"
    VN = "vn.biggo.com"
    TH = "biggo.co.th"
    PH = "ph.biggo.com"
    MY = "my.biggo.com"
    IN = "biggo.co.in"
    US = "biggo.com"
    TW = "biggo.com.tw"
    HK = "biggo.hk"
    JP = "biggo.jp"
    SG = "biggo.sg"


REGION_DOMAIN_MAP: dict[Regions, Domains] = {
    Regions.ID: Domains.ID,
    Regions.VN: Domains.VN,
    Regions.TH: Domains.TH,
    Regions.PH: Domains.PH,
    Regions.MY: Domains.MY,
    Regions.IN: Domains.IN,
    Regions.US: Domains.US,
    Regions.TW: Domains.TW,
    Regions.HK: Domains.HK,
    Regions.JP: Domains.JP,
    Regions.SG: Domains.SG,
}


class BigGoMCPSetting(BaseSettings):
    """
    BigGo MCP Server settings
    """

    model_config = SettingsConfigDict(env_prefix="BIGGO_MCP_SERVER_")

    region: Regions = Regions.TW

    client_id: str | None = None
    client_secret: str | None = None

    log_level: LogLevel = LogLevel.INFO

    es_proxy_url: str = "https://api.biggo.com/api/v1/mcp-es-proxy/"
    es_verify_certs: bool = True

    auth_token_url: str = "https://api.biggo.com/auth/v1/token"
    auth_verify_certs: bool = True

    sse_port: int = 9876

    server_type: Literal["stdio", "sse"] = "stdio"

    short_url_endpoint: str | None = None

    @field_validator("region", mode="before")
    @classmethod
    def _validate_region(cls, value: Any) -> Regions:
        # Make region case insensitive
        if isinstance(value, str):
            return Regions(value.upper())
        else:
            return value

    @property
    def domain(self) -> Domains:
        return REGION_DOMAIN_MAP[self.region]

    @property
    def graph_language(self) -> GraphLanguage:
        if self.region == Regions.TW:
            return GraphLanguage.TW
        else:
            return GraphLanguage.EN



from biggo_mcp_server.types.api_ret.product_search import ListItem, ProductSearchAPIRet
from biggo_mcp_server.types.setting import Domains


def test_product_search_api_ret():
    data = ProductSearchAPIRet(
        list=[
            ListItem(
                oid="123",
                title="test",
                price=100,
                affurl="/r/some-link",
            )
        ]
    )

    assert data.list[0].url == ""
    data.generate_r_link(Domains.TW)

    assert data.list[0].url == "https://biggo.com.tw/r/some-link"



from aiohttp import ClientSession

from biggo_mcp_server.lib.utils import generate_short_url
from ..types.api_ret.product_search import ProductSearchAPIRet
from ..types.setting import BigGoMCPSetting
from logging import getLogger

logger = getLogger(__name__)


class ProductSearchService:
    def __init__(self, setting: BigGoMCPSetting):
        self._setting = setting

    async def search(self, query: str) -> ProductSearchAPIRet:
        url = f"https://api.biggo.com/api/v1/spa/search/{query}/product"

        headers = {
            "Content-Type": "application/json",
            "site": self._setting.domain.value,
            "region": self._setting.region.value.lower(),
        }
        logger.debug("product search, url: %s, headers: %s", url, headers)

        async with ClientSession() as session:
            async with session.get(url, headers=headers) as resp:
                if resp.status >= 400:
                    err_msg = f"product search api error: {await resp.text()}"
                    logger.error(err_msg)
                    raise ValueError(err_msg)

                data = ProductSearchAPIRet.model_validate(await resp.json())

        data.generate_r_link(self._setting.domain)

        if self._setting.short_url_endpoint is not None:
            all_urls = data.get_all_urls()
            url_map = await generate_short_url(
                list(all_urls), self._setting.short_url_endpoint
            )
            data.replace_urls(url_map)

        return data



from logging import getLogger
from typing import Annotated
from mcp.server.fastmcp import Context
from pydantic import Field
from ..types.responses import PriceHisotryGraphToolResponse, PriceHistoryToolResponse
from ..services.price_history import DAYS, PriceHistoryService
from ..lib.utils import get_setting

logger = getLogger(__name__)


def price_history_graph(
    ctx: Context,
    history_id: Annotated[
        str,
        Field(
            description="""
              Product History ID - Get from 'history_id' field in product_search results
              """,
            examples=["tw_pmall_rakuten-nwdsl_6MONJRBOO", "tw_pec_senao-1363332"],
        ),
    ],
) -> str:
    """Link That Visualizes Product Price History"""

    logger.info("price history graph, history_id: %s", history_id)

    setting = get_setting(ctx)
    service = PriceHistoryService(setting)
    url = service.graph_link(history_id)

    return PriceHisotryGraphToolResponse(
        price_history_graph=f"![Price History Graph]({url})"
    ).slim_dump()


async def price_history_with_history_id(
    ctx: Context,
    history_id: Annotated[
        str,
        Field(
            description="""
              Product History ID - Get from 'history_id' field in product_search results
              """,
            examples=["tw_pmall_rakuten-nwdsl_6MONJRBOO", "tw_pec_senao-1363332"],
        ),
    ],
    days: Annotated[
        DAYS, Field(description="History range, must be one of 90, 80, 365, 730")
    ] = "90",
) -> str:
    """Product Price History With History ID"""

    logger.info(
        "price history with history id, history_id: %s, days: %s", history_id, days
    )

    setting = get_setting(ctx)
    service = PriceHistoryService(setting)
    ret = await service.history_with_history_id(history_id=history_id, days=days)
    if ret is None:
        return "No price history found"

    return PriceHistoryToolResponse(
        price_history_description=ret.description,
        price_history_graph=f"![Price History Graph]({ret.graph_link})",
    ).slim_dump()


async def price_history_with_url(
    ctx: Context,
    url: Annotated[str, Field(description="Product URL")],
    days: Annotated[
        DAYS,
        Field(description="History range, must be one of 90, 80, 365, 730"),
    ] = "90",
) -> str:
    """Product Price History With URL"""

    logger.info("price history with url, url: %s, days: %s", url, days)

    setting = get_setting(ctx)
    service = PriceHistoryService(setting)
    ret = await service.history_with_url(url=url, days=days)
    if ret is None:
        return "No price history found"

    return PriceHistoryToolResponse(
        price_history_description=ret.description,
        price_history_graph=f"![Price History Graph]({ret.graph_link})",
    ).slim_dump()



from pydantic import ValidationError
import pytest
from biggo_mcp_server.types.setting import BigGoMCPSetting
from .helper import *


def test_setting_region_case_insensitive():
    env_key = "BIGGO_MCP_SERVER_REGION"

    # upper case
    with set_temp_env({env_key: "TW"}):
        BigGoMCPSetting()

    # lower case
    with set_temp_env({env_key: "tw"}):
        BigGoMCPSetting()

    # random stuff
    with set_temp_env({env_key: "not-a-region"}):
        with pytest.raises(ValidationError):
            BigGoMCPSetting()



from biggo_mcp_server.types.api_ret.product_search import ListItem, ProductSearchAPIRet
from biggo_mcp_server.types.responses import (
    ProductSearchToolResponse,
    SpecSearchToolResponse,
)


def test_product_search_resp():
    resp = ProductSearchToolResponse(product_search_result=ProductSearchAPIRet(list=[]))

    assert resp.reason is not None

    resp = ProductSearchToolResponse(
        product_search_result=ProductSearchAPIRet(
            list=[
                ListItem(
                    oid="123",
                    title="test",
                    price=100,
                    affurl="/r/some-link",
                )
            ]
        )
    )
    assert resp.reason is None


def test_spec_search_resp():
    resp = SpecSearchToolResponse(hits=[])

    assert resp.reason is not None

    resp = SpecSearchToolResponse(hits=[{"some": "data", "_source": {}}])

    assert resp.reason is None
